@@ -61,6 +61,8 @@
 #include <linux/ioport.h>
 #include <linux/init.h>
 
+#include <linux/spinlock.h>
+
 #include "mga_vid.h"
 
 #ifdef CONFIG_MTRR
@@ -286,6 +288,8 @@ typedef struct mga_card_s {
 	unsigned char colkey_mask[4];
 
 	int next_frame;
+
+	spinlock_t lock;
 } mga_card_t;
 
 static unsigned int mga_cards_num = 0;
@@ -653,53 +657,71 @@ static int mga_vid_set_config(mga_card_t * card)
 	return 0;
 }
 
-static int mga_vid_ioctl(struct inode *inode, struct file *file,
-			 unsigned int cmd, unsigned long arg)
+static long mga_vid_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
 {
-	int frame, result;
+	int frame, result = 0;
 	uint32_t tmp;
 	mga_card_t *card = (mga_card_t *) file->private_data;
 
+	spin_lock(&card->lock);
+
 	switch (cmd) {
+	case MGA_VID_GET_VERSION:
+		tmp = MGA_VID_VERSION;
+		if (copy_to_user((uint32_t *) arg, &tmp, sizeof(uint32_t))) {
+			printk(KERN_ERR "mga_vid: failed copy %p to userspace %p\n", &tmp, (uint32_t *) arg);
+			result = -EFAULT;
+			goto ioctl_out;
+		}
+		break;
+
 	case MGA_VID_CONFIG:
 		if (copy_from_user (&card->config, (mga_vid_config_t *) arg, sizeof(mga_vid_config_t))) {
 			printk(KERN_ERR "mga_vid: failed copy from userspace\n");
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 		// return some information back to user space in case we hit an error
 		card->config.card_type = card->card_type;
 		card->config.ram_size = card->ram_size;
+		card->config.capabilities = 0;
 		if (copy_to_user ((mga_vid_config_t *) arg, &card->config, sizeof(mga_vid_config_t))) {
 			printk(KERN_ERR "mga_vid: failed copy to userspace\n");
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 
 		if (card->config.version != MGA_VID_VERSION) {
 			printk(KERN_ERR "mga_vid: incompatible version! driver: %X  requested: %X\n", MGA_VID_VERSION, card->config.version);
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 
 		if (card->config.frame_size == 0 || card->config.frame_size > 1920 * 1080 * 2) {
 			printk(KERN_ERR "mga_vid: illegal frame_size: %d\n", card->config.frame_size);
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 
 		if (card->config.num_frames < 1 || card->config.num_frames > 4) {
 			printk(KERN_ERR "mga_vid: illegal num_frames: %d\n", card->config.num_frames);
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 
 		if ((card->config.src_width > 1024 || card->config.src_height > 1024)
 		    && (card->card_type != MGA_G550 && card->config.src_height > 1024)) {
 			printk(KERN_ERR "mga_vid: overlay sizes bigger than 1024x1024 are not supported with this card type, rescale the picture in software\n");
 
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 
 		card->src_base = (card->ram_size * 0x100000 - card->config.num_frames * card->config.frame_size - card->top_reserved);
 		if (card->src_base < 0) {
 			printk(KERN_ERR "mga_vid: not enough memory for frames!\n");
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 		card->src_base &= (~0xFFFF);	// 64k boundary
 #ifdef MP_DEBUG
@@ -707,13 +729,14 @@ static int mga_vid_ioctl(struct inode *inode, struct file *file,
 #endif
 		if (copy_to_user ((mga_vid_config_t *) arg, &card->config, sizeof(mga_vid_config_t))) {
 			printk(KERN_ERR "mga_vid: failed copy to userspace\n");
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 
 		result = mga_vid_set_config(card);
 		if (!result)
 			card->configured = 1;
-		return result;
+		goto ioctl_out;
 		break;
 
 	case MGA_VID_ON:
@@ -739,9 +762,10 @@ static int mga_vid_ioctl(struct inode *inode, struct file *file,
 		break;
 
 	case MGA_VID_FSEL:
-		if (copy_from_user(&frame, (int *)arg, sizeof(int))) {
+		if (copy_from_user(&frame, (int *)arg, sizeof(uint32_t))) {
 			printk(KERN_ERR "mga_vid: FSEL failed copy from userspace\n");
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 
 		mga_vid_frame_sel(card, frame);
@@ -754,7 +778,8 @@ static int mga_vid_ioctl(struct inode *inode, struct file *file,
 
 		if (copy_to_user((uint32_t *) arg, &tmp, sizeof(uint32_t))) {
 			printk(KERN_ERR "mga_vid: failed copy %p to userspace %p\n", &tmp, (uint32_t *) arg);
-			return (-EFAULT);
+			result = -EFAULT;
+			goto ioctl_out;
 		}
 		break;
 
@@ -769,10 +794,13 @@ static int mga_vid_ioctl(struct inode *inode, struct file *file,
 
 	default:
 		printk(KERN_ERR "mga_vid: Invalid ioctl\n");
-		return (-EINVAL);
+		result = -EINVAL;
+		goto ioctl_out;
 	}
 
-	return 0;
+	ioctl_out:
+	spin_unlock(&card->lock);
+	return result;
 }
 
 static void cards_init(mga_card_t * card, struct pci_dev *dev, int card_number, int card_type);
@@ -868,7 +896,7 @@ static ssize_t mga_vid_read(struct file *file, char *buf, size_t count, loff_t *
 	if (*ppos >= card->param_buff_len)
 		return 0;
 	size = min(count, card->param_buff_len - (size_t) (*ppos));
-	memcpy(buf, card->param_buff, size);
+	copy_to_user(buf, card->param_buff, size);
 	*ppos += size;
 	return size;
 }
@@ -876,39 +904,71 @@ static ssize_t mga_vid_read(struct file *file, char *buf, size_t count, loff_t *
 static ssize_t mga_vid_write(struct file *file, const char *buf, size_t count, loff_t * ppos)
 {
 	mga_card_t *card = (mga_card_t *) file->private_data;
+	char * buffer;
 
-	if (memcmp(buf, PARAM_BRIGHTNESS, min(count, strlen(PARAM_BRIGHTNESS))) == 0) {
+	if(count >= PAGE_SIZE)
+	{
+		// only one parameter is handled at a time,
+		// so writing more than one page does not
+		// make sense anyways
+                printk(KERN_ERR "mga_vid: writing more than %lu bytes to the device file\n",PAGE_SIZE);
+		return -EFBIG;
+	}
+
+	// allocate memory and copy data into our own buffer
+	buffer = kmalloc(count, GFP_KERNEL);
+	
+	if(!buffer)
+	{
+                printk(KERN_ERR "mga_vid: couln't allocate buffer\n");
+		return -EIO;
+	}
+
+	memset(buffer, 0, count);
+	// copy_from_user returns the amount of data _still_ to be copied
+	if(copy_from_user(buffer, buf, count))
+	{
+		// something strange happend, bail out
+                printk(KERN_ERR "mga_vid: couldn't copy data from user space\n");
+		kfree(buffer);
+		return -EIO;
+	}
+
+	if (memcmp(buffer, PARAM_BRIGHTNESS, min(count, strlen(PARAM_BRIGHTNESS))) == 0) {
 		short brightness;
-		brightness = simple_strtol(&buf[strlen(PARAM_BRIGHTNESS)], NULL, 10);
+		brightness = simple_strtol(&buffer[strlen(PARAM_BRIGHTNESS)], NULL, 10);
 		if (brightness > 127 || brightness < -128) {
 			brightness = 0;
 		}
 //              printk(KERN_DEBUG "mga_vid: brightness modified ( %d ) \n",brightness);
 		card->brightness = brightness;
 	} else
-	    if (memcmp(buf, PARAM_CONTRAST, min(count, strlen(PARAM_CONTRAST))) == 0) {
+	    if (memcmp(buffer, PARAM_CONTRAST, min(count, strlen(PARAM_CONTRAST))) == 0) {
 		short contrast;
-		contrast = simple_strtol(&buf[strlen(PARAM_CONTRAST)], NULL, 10);
+		contrast = simple_strtol(&buffer[strlen(PARAM_CONTRAST)], NULL, 10);
 		if (contrast > 127 || contrast < -128) {
 			contrast = 0;
 		}
 //              printk(KERN_DEBUG "mga_vid: contrast modified ( %d ) \n",contrast);
 		card->contrast = contrast;
 	} else
-	 if (memcmp(buf, PARAM_BLACKIE, min(count, strlen(PARAM_BLACKIE))) == 0) {
+	 if (memcmp(buffer, PARAM_BLACKIE, min(count, strlen(PARAM_BLACKIE))) == 0) {
 		short blackie;
-		blackie = simple_strtol(&buf[strlen(PARAM_BLACKIE)], NULL, 10);
+		blackie = simple_strtol(&buffer[strlen(PARAM_BLACKIE)], NULL, 10);
 //              printk(KERN_DEBUG "mga_vid: shadow mode: ( %d ) \n",blackie);
 		card->regs.blackie = (blackie > 0) ? 1 : 0;
 	} else
 		count = -EIO;
 	// TODO: reset settings
+	
+	kfree(buffer);
 	return count;
 }
 
 static int mga_vid_mmap(struct file *file, struct vm_area_struct *vma)
 {
 	mga_card_t *card = (mga_card_t *) file->private_data;
+	spin_lock(&card->lock);
 
 #ifdef MP_DEBUG
 	printk(KERN_DEBUG "mga_vid: mapping video memory into userspace\n");
@@ -916,6 +976,7 @@ static int mga_vid_mmap(struct file *file, struct vm_area_struct *vma)
 
 	if (!card->configured) {
 		printk(KERN_ERR "mga_vid: card is not configured, cannot mmap\n");
+		spin_unlock(&card->lock);
 		return (-EAGAIN);
 	}
 	//FIXME: should we set VM_IO too ? remap_pfn_range does it for us,
@@ -926,9 +987,11 @@ static int mga_vid_mmap(struct file *file, struct vm_area_struct *vma)
 			    (card->mem_base + card->src_base) >> PAGE_SHIFT,
 			    vma->vm_end - vma->vm_start, vma->vm_page_prot)) {
 		printk(KERN_ERR "mga_vid: error mapping video memory\n");
+		spin_unlock(&card->lock);
 		return (-EAGAIN);
 	}
 
+	spin_unlock(&card->lock);
 	return 0;
 }
 
@@ -942,6 +1005,7 @@ static int mga_vid_release(struct inode *inode, struct file *file)
 #endif
 
 	card = (mga_card_t *) file->private_data;
+	spin_lock(&card->lock);
 
 	card->vid_src_ready = 0;
 	card->regs.besctl &= ~1;
@@ -949,6 +1013,7 @@ static int mga_vid_release(struct inode *inode, struct file *file)
 //      card->config.colkey_on=0; //!!!
 	mga_vid_write_regs(card, 1);
 	card->vid_in_use = 0;
+	spin_unlock(&card->lock);
 	return 0;
 }
 
@@ -971,11 +1036,16 @@ static int mga_vid_open(struct inode *inode, struct file *file)
 
 	card = (mga_card_t *) file->private_data;
 
+	spin_lock(&card->lock);
 	if (card->vid_in_use == 1)
+	{
+		spin_unlock(&card->lock);
 		return (-EBUSY);
+	}
 
 	card->vid_in_use = 1;
 	card->configured = 0;
+	spin_unlock(&card->lock);
 	return 0;
 }
 
@@ -983,7 +1053,8 @@ static struct file_operations mga_vid_fops = {
       llseek:mga_vid_lseek,
       read:mga_vid_read,
       write:mga_vid_write,
-      ioctl:mga_vid_ioctl,
+      unlocked_ioctl:mga_vid_ioctl,
+      compat_ioctl:mga_vid_ioctl,
       mmap:mga_vid_mmap,
       open:mga_vid_open,
       release:mga_vid_release
@@ -993,6 +1064,7 @@ static void cards_init(mga_card_t * card, struct pci_dev *dev, int card_number,
 {
 	unsigned int card_option;
 
+	spin_lock_init(&card->lock);
 	card->pci_dev = dev;
 
 	card->card_type = card_type;
@@ -1094,7 +1166,7 @@ static int mga_vid_initialize(void)
 
 				mga_vid_class = class_create(THIS_MODULE, "mga_vid");
 				for (i = 0; i < mga_cards_num; i++) {
-					class_device_create(mga_vid_class, NULL, MKDEV(mga_vid_major, i), NULL, "mga_vid%d", i);
+					device_create(mga_vid_class, NULL, MKDEV(mga_vid_major, i), NULL, "mga_vid%d", i);
 				}
 
 				printk(KERN_INFO "mga_vid: using major: %d (assigned or default!)\n", major);
@@ -1143,7 +1215,7 @@ static void mga_cleanup_module(void)
 	}
 
 	for (i = 0; i < mga_cards_num; i++) {
-		class_device_destroy(mga_vid_class, MKDEV(mga_vid_major, i));
+		device_destroy(mga_vid_class, MKDEV(mga_vid_major, i));
 	}
 	class_destroy(mga_vid_class);
 
